#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>
#include <sstream>

#include "rng.h"
#include "tree.h"
#include "info.h"
#include "funs.h"
#include "calibd.h"
#include "calirotp.h"

using std::cout;
using std::endl;

int main(int argc, char** argv)
{
#ifndef SILENT
   cout << "\n*****Into bart main\n";
#endif

	//--------------------------------------------------
	// Load up the settings for running the model
   	//
	// settings.txt has the following format (1 var per line)
	// "x", "xf", "y", "yf"
	// nu, nuf, lambda, lambdaf,
	// tau0, mu0, tauc, muc, taud, mud,
	// m, numcuts, pbirth, alpha, beta
	// nd, thin, burn, pbd, pert.alpha

	// First load the filenames to read the datasets from.
	std::string xroot;
	std::string xfroot;
	std::string yroot;
	std::string yfroot;
	std::ifstream sfi("settings.txt");

	sfi >> xroot >> xfroot >> yroot >> yfroot;
	cout << "File  x: " << xroot << ".txt" << endl;
	cout << "File xf: " << xfroot << ".txt" << endl;
	cout << "File  y: " << yroot << ".txt" << endl;
	cout << "File yf: " << yfroot << ".txt" << endl;

	double lambda;
	double lambdaf;
	double nu;
	double nuf;

	sfi >> nu >> nuf >> lambda >> lambdaf;
	cout << "nu  = " << nu << endl;
	cout << "nuf = " << nuf << endl;
	cout << "lambda  = " << lambda << endl;
	cout << "lambdaf = " << lambdaf << endl;

	pinfo pi;

	sfi >> pi.tau0 >> pi.mu0 >> pi.tauc >> pi.muc >> pi.taud >> pi.mud;
	cout << "tau0 = " << pi.tau0 << endl;
	cout << "mu0  = " << pi.mu0 << endl;
	cout << "tauc = " << pi.tauc << endl;
	cout << "muc  = " << pi.muc << endl;
	cout << "taud = " << pi.taud << endl;
	cout << "mud  = " << pi.mud << endl;

	size_t m; //number of trees in BART sum
	size_t numcuts; //number of cutpoints except for imdim.
	sfi >> m >> numcuts;
	sfi >> pi.pb; // prob of birth given birth/death move
	sfi >> pi.alpha; //0.95 prior prob a bot node splits is alpha/(1+d)^beta, d is depth of node
	sfi >> pi.beta;  //2 for bart means it is harder to build big trees.

	cout << "num trees m   = " << m << endl;
	cout << "num cutpoints = " << numcuts << endl;
	cout << "prob birth = " << pi.pb << endl;
	cout << "alpha = " << pi.alpha << endl;
	cout << "beta = " << pi.beta << endl;

	size_t nd;   //number of mcmc iterations
	size_t thin; //save a draw ever thin iterations after burn-in.
	size_t burn; //number of mcmc iterations called burn-in
	sfi >> nd >> thin >> burn;
	sfi >> pi.pbd;		 //prob of birth/death move, otherwise we do rotate.
	sfi >> pi.pertalpha; //for perturb proposal, something like .05 or 0.1 usually. 
						 //use this to tune acceptance rate of tree moves.

	cout << "num MCMC draws nd = " << nd << endl;
	cout << "save draws every thin = " << thin << endl;
	cout << "discarded draws burn = " << burn << endl;
	cout << "prob of birth/death move pbd = " << pi.pbd << endl;
	cout << "perturb proposal alpha = " << pi.pertalpha << endl;

	sfi.close(); //end of loading settings.


	//--------------------------------------------------
	//random number generation
	uint seed=time(0); // set it to 99 for debugging purposes only.
	RNG gen(seed);
	cout << "RNG seed is " << seed << "\n";


	//--------------------------------------------------
	//read in data
	//read y NOTE: we assume data is already centered at 0.
	sinfo allys;
	double xtemp,ytemp; //used to read in x and y without knowing how many there are.
	double sy=0.0, sy2=0.0, sf=0.0, sf2=0.0;
	unsigned int tempn=0, tempnf=0;
	size_t n=0,nn=0;
	size_t nf=0,nnf=0;
	double miny=INFINITY;
	double maxy=-INFINITY;
	double myminy=INFINITY;
	double mymaxy=-INFINITY;
	double minf=INFINITY;
	double maxf=-INFINITY;
	double myminf=INFINITY;
	double mymaxf=-INFINITY;
	std::vector<double> y;   //storage for y on the slaves
	std::vector<double> yf;  //storage for yf on the slaves
	std::stringstream yfss;
	std::stringstream yffss;
	std::string yfs;

	// First load the yf's
	yffss << yfroot << ".txt";
	yfs=yffss.str();
	std::ifstream yfi(yfs.c_str()); //file to read y from for this slave
	
	while(yfi >> ytemp) {
		yf.push_back(ytemp);
		if(ytemp<myminf) myminf=ytemp;
		if(ytemp>mymaxf) mymaxf=ytemp;
		sf+=ytemp;
		sf2+=ytemp*ytemp;
	}
	tempnf=yf.size();
	nnf=(size_t)tempnf;
	if(nnf<1) {
		cout << "error nnf<1\n";
		return 1;
	}

	// And now load the y's
	yfss << yroot << ".txt";
	yfs=yfss.str();
	std::ifstream yi(yfs.c_str()); //file to read y from for this slave

	while(yi >> ytemp) {
		y.push_back(ytemp);
		if(ytemp<myminy) myminy=ytemp;
		if(ytemp>mymaxy) mymaxy=ytemp;
		sy+=ytemp;
		sy2+=ytemp*ytemp;
	}
	tempn=y.size();
	nn=(size_t)tempn;
	if(nn<1) {
		cout << "error nn<1\n";
		return 1;
	}

#ifndef SILENT
	cout << ": nnf=" << nnf << " nn=" << nn << " sy=" << sy << " sy2=" << sy2 << " myminy=" << myminy << " mymaxy=" << mymaxy << " sf=" << sf << " sf2=" << sf2 << " myminf=" << myminf << " mymaxf=" << mymaxf << endl;
#endif

	miny=myminy;
	maxy=mymaxy;
	minf=myminf;
	maxf=mymaxf;
	allys.sy=sy;
	allys.sy2=sy2;
	allys.sf=sf;
	allys.sf2=sf2;
	n=(size_t) tempn;
	nf=(size_t) tempnf;

#ifndef SILENT
	cout << ": nf=" << nf << " nnf=" << nnf << " n=" << n << " nn=" << nn << " allys.sy=" << allys.sy << " allys.sy2=" << allys.sy2 << " miny=" << miny << " maxy=" << maxy << " allys.sf=" << allys.sf << " allys.sf2=" << allys.sf2 << " minf=" << minf << " maxf=" << maxf << endl;
#endif

	if(n<1) {
		cout << "error in loading data: n<1\n";
		return 1;
	}
	if(nf<1) {
		cout << "error in loading data: nf<1\n";
		return 1;
	}

	allys.n=n;
	allys.nf=nf;
	double ybar = allys.sy/n;
	double ybarf = allys.sf/nf;
    double shat = sqrt((allys.sy2-n*ybar*ybar)/(n-1));
    double sfhat = sqrt((allys.sf2-nf*ybarf*ybarf)/(nf-1));

#ifndef SILENT
   cout << "ybar,shat  : " << ybar << ", " << shat <<  endl;
   cout << "ybarf,sfhat: " << ybarf << ", " << sfhat << endl;
#endif

	//read x
	//the n*p numbers for x are stored as the p for first obs, then p for second, and so on.
	unsigned int myp=0, mypf=0;
	size_t p=0, pf=0;
	std::vector<double> x;
	std::vector<double> xf, xft;

	// First load up the xf's corresponding to the field observations
	std::stringstream xffss;
	std::string xffs;
	xffss << xfroot << ".txt";
	xffs=xffss.str();
	std::ifstream xfi(xffs.c_str());  //file to read x from
	while(xfi >> xtemp) 
	   xf.push_back(xtemp);
	mypf = (unsigned int)xf.size()/nnf;
	pf = (size_t) mypf;
	if(xf.size() != nnf*pf) {
			cout << "error: input xf file has wrong number of values\n"; 
			return 1;
	}

	// and then load up the x's.
	std::stringstream xfss;
	std::string xfs;
	xfss << xroot << ".txt";
	xfs=xfss.str();
	std::ifstream xxi(xfs.c_str());  //file to read x from
	while(xxi >> xtemp) 
	   x.push_back(xtemp);
	myp = (unsigned int)x.size()/nn;
	p = (size_t) myp;
	if(x.size() != nn*p) {
		cout << "x.size=" << x.size() << endl;
		cout << "myp=" << myp << endl;
		cout << "p  =" << p << endl;
		cout << " nn=" << nn << endl;
			cout << "error: input x file has wrong number of values\n"; 
			return 1;
	}
	if(p != pf) {
		cout << "error: input files xf and x don't match on number of dimensions\n";
		return 1;
	}

   //x for predictions
   dinfo dip; //data information for prediction
   dip.n=0;
   dip.nf=0;
   //Currently no out-of-sample prediction
#ifndef SILENT
   cout << "\nx for prediction read in:\n";
   cout << "n: " << dip.n << endl;
#endif
   if(dip.n) {
#ifndef SILENT
      cout << "first row: " <<  dip.x[0] << " ...  " << dip.x[p-1] << endl;
      cout << "last row: " << dip.x[(dip.n-1)*p] << " ...  " << dip.x[dip.n*p-1] << endl;
#endif
   }

  	// Load up the calibration parameter dimension indices
	std::stringstream cfss;
	std::string cfs;
	unsigned int ctemp, ncaldims;
	std::vector<unsigned int> caldims;
	cfss << "caldims.txt";
		cfs=cfss.str();
#ifndef SILENT
	cout << ": loading calibration parameter dimensions from " << cfs << endl;
#endif
		std::ifstream cf(cfs.c_str()); //file to read caldims from for this slave
	while(cf >> ctemp) {
		caldims.push_back(ctemp-1);
		if(ctemp<0 || ctemp>p)
		{
			cout << "error: calibration dimensions is outside range of number of predictors p" << endl;
			return 1;
		}
	}
	ncaldims=caldims.size();
	std::vector<double> calests(ncaldims);
#ifndef SILENT
	cout << ": loaded " << ncaldims << " calibration dimensions: ";
	for(size_t i=0;i<ncaldims;i++)
		cout << caldims[i] << " ";
	cout << endl;
#endif

	// Load up the model/field indicator variable Im
	std::stringstream imfss;
	std::string imfs;
	unsigned int imdim;
	imfss << "imdim.txt";
	imfs=imfss.str();
#ifndef SILENT
	cout << ": loading model/field indicator variable Im from " << imfs << endl;
#endif
	std::ifstream imf(imfs.c_str()); //file to read Im from
	imf >> imdim;
	imdim--;
	if(imdim<0 || imdim>p)
	{
		cout << "error: Imdim dimension is outside the range of number of predictors p" << endl;
		return 1;
	}
#ifndef SILENT
	cout << ": loaded Imdim dimension Im=" << imdim << endl;
#endif


	//--------------------------------------------------
	//x cutpoints
	std::vector<double> minx(p,INFINITY);
	std::vector<double> maxx(p,-INFINITY);
	xinfo xi;
	std::vector<size_t> nc(p,numcuts); //default number of cutpoints per covariate
	for(size_t i=0;i<ncaldims;i++)
		nc[caldims[i]]=numcuts;  //can alter num cutpoints for calibration parameters
	nc[imdim]=1; //for imdim there is only 1 cutpoint as it's a binary indicator variable
	makeminmax(p,nn,&x[0],minx,maxx);
	minx[imdim]=0; maxx[imdim]=1;  // we set this manually because makeminmax() is finding min/max based on X not X AND Xf.
	makexinfominmax(p,xi,nc,minx,maxx); // make the xinfo xi struct


	//--------------------------------------------------
	//trees
	std::vector<tree> t(m);
	for(size_t i=0;i<m;i++) t[i].setm((n*ybar+nf*ybarf)/(n+nf)/m); //if you sum the fit over the trees you get the fit.

	//--------------------------------------------------
	//prior and mcmc - additional things we didnt'set earlier from settings.txt
	pi.sigma=shat;
	pi.sigmaf=sfhat;
	pi.imdim=imdim;

#ifndef SILENT
   cout << "\nalpha, beta: " << pi.alpha << ", " << pi.beta << endl;
   cout << "sigma, tau: " << pi.sigma << ", " << pi.tau << endl;
   cout << "sigmaf, tauf: " << pi.sigmaf << ", " << pi.tauf << endl;
#endif

	//--------------------------------------------------
	//read the change-of-variable proposal probability matrix
	//this is a p x p matrix
	std::ifstream cvpmf("chgvarmat.txt");  //file to read x from
	double cvptemp;
	for(size_t i=0;i<p;i++) {
	  std::vector<double> cvvtemp;
	  for(size_t j=0;j<p;j++) {
	        cvpmf >> cvptemp;
	    cvvtemp.push_back(cvptemp);
	  }
	  pi.cvpm.push_back(cvvtemp);
	}

	cout << "Change of Variable proposal probability matrix loaded:" << endl;
	for(size_t i=0;i<p;i++) {
		for(size_t j=0;j<p;j++)
			cout << "(" << i << "," << j << ")" << pi.cvpm[i][j] << "  ";
	    cout << endl;
	}


	//--------------------------------------------------
	//dinfo
	dinfo di;
	double *ftemp=NULL;
	double *fftemp=NULL;
	double *r=NULL;
	double *rf=NULL;
	double *allfit=NULL;
	double *allfitf=NULL;
	
	allfit = new double[n]; //sum of fit of all trees
	for(size_t i=0;i<n;i++) allfit[i]=ybar;
	allfitf= new double[nf];
	for(size_t i=0;i<nf;i++) allfitf[i]=ybarf;
	r = new double[n]; //y-(allfit-ftemp) = y-allfit+ftemp
	rf= new double[nf];
	ftemp = new double[n]; //fit of current tree
	fftemp= new double[nf];
	di.n=n; di.p=p; di.x = &x[0]; di.y=r; //the y for each draw will be the residual
	di.nf=nf; di.xf = &xf[0]; di.yf=rf;


	//--------------------------------------------------
	// initialize the calibration dimensions to be at the midpoints
	for(size_t i=0;i<ncaldims;i++)
	   calests[i]=0.5; //initialize to midpoint -- we ASSUME variables are in [0,1]!
	for(size_t i=0;i<ncaldims;i++)
	   for(size_t j=0;j<di.nf;j++)
	       xf[di.p*j + caldims[i]]=calests[i];
	xft=xf; //a copy of xf, we'll need it for doing the acc/rej steps later on.
	pi.caldims=caldims;
	pi.calests=calests;
	pi.ncaldims=ncaldims;
	pi.taccept=0;

	std::vector<size_t>calacc(ncaldims);
	std::vector<double>calpropscale(ncaldims);
	pi.calacc=calacc;
	pi.calpropscale=calpropscale;
	double calrate;
	size_t lastadapt=0;
	for(size_t i=0;i<ncaldims;i++) {
		pi.calacc[i]=0;
		pi.calpropscale[i]=0.1;  // default stepwidth for proposal.
	}

	cout << ": caldims=";
	for(size_t i=0;i<ncaldims;i++)
		cout << " " << pi.caldims[i];
	cout << endl;
	cout << ": calests=";
	for(size_t i=0;i<ncaldims;i++)
		cout << " " << pi.calests[i];
	cout << endl;
	cout << ": xf[1..p]=";
	for(size_t i=0;i<di.p;i++)
	    cout << " " << xf[i];
	cout << endl;


	//--------------------------------------------------
	//storage for ouput
	//in sample fit
	double *pmean;
	double *pfmean;
	//for sigma draw
	double rss; //residual sum of squares
	double restemp; //a residual
	double rssf;
	//out of sample fit
	double* ppredmean=0; //posterior mean for prediction
	double* fpredtemp=0; //temporary fit vector to compute prediction

	std::ofstream bsdf("bart-sd.txt"); //note that we write all burn+nd draws to this file.
	std::ofstream bsdff("bart-sdf.txt");
	std::ofstream bcpf("bart-caldims.txt");  //for the calibration parameter estimates.

	pmean = new double[n]; //posterior mean of in-sample fit, sum draws,then divide
		for(size_t i=0;i<n;i++) pmean[i]=0.0;
		pfmean= new double[nf];
		for(size_t i=0;i<nf;i++) pfmean[i]=0.0;
		if(dip.n) {
	  	ppredmean = new double[dip.n];
	  	fpredtemp = new double[dip.n];
	  	for(size_t i=0;i<dip.n;i++) ppredmean[i]=0.0;
		}


	//--------------------------------------------------
	//mcmc
	//Setup saving of the fitted trees
	std::ofstream treef("trees.txt");
	treef << xi << endl; //cutpoints
	treef << m << endl;  //number of trees
	treef << p << endl;  //dimension of x's
	treef << (int)(nd/thin) << endl; //number of draws from posterior that will be saved

	#ifndef SILENT
	cout << "\nMCMC:\n";
	#endif
	for(size_t i=0;i<(nd+burn);i++)
	{
		#ifndef SILENT
		if(i%100==0) cout << "i: " << i << endl;
		#endif


		//draw calibration parameters
		makefieldpred(di,xi,t,allfitf);
		calidrcp(t,xi,di,pi,xf,xft,yf,allfitf,y,allfit,gen);

		if(i<burn && (i+1)%250==0) {
			cout << "i=" << i << endl;
			cout << "         acc.rate" << endl;
			for(size_t j=0;j<ncaldims;j++) {
				pi.calacc[j]++; // a hack, gets rid of rate=0 case.
				calrate=((double)(pi.calacc[j]))/((double)(i-lastadapt));
				if(calrate<0.35 || calrate>0.55) pi.calpropscale[j]*=calrate/0.44;
				cout << "param " << j+1 << ": " << calrate << "   (ests=" << pi.calests[j] << ", calacc=" << pi.calacc[j] << ", calpropscale=" << pi.calpropscale[j] << ") " << endl;
				pi.calacc[j]=0;
			}
			cout << endl;
			lastadapt=i;
		}

		//draw trees
		for(size_t j=0;j<m;j++) 
		{
			fit(t[j],xi,di,ftemp);
			fitfield(t[j],xi,di,fftemp);
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<n;k++)
			{
				allfit[k] = allfit[k]-ftemp[k];
				r[k] = y[k]-allfit[k];
			}
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<nf;k++)
			{
				allfitf[k] = allfitf[k]-fftemp[k];
				rf[k] = yf[k] - allfitf[k];
			}

		 	if(gen.uniform()>pi.pbd) {
				tree::tree_p tnew;
				tnew=new tree(t[j]); //copy of current to make life easier upon rejection
				calirotp(tnew,t[j],xi,di,pi,gen);
				delete tnew;
		 	} else {
				calibd(t[j],xi,di,pi,gen);
			}

			calichgv(t[j],xi,di,pi,gen);
			calidrmu(t[j],xi,di,pi,gen);
			fit(t[j],xi,di,ftemp);
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<n;k++) allfit[k] += ftemp[k];
			fitfield(t[j],xi,di,fftemp);
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<nf;k++) allfitf[k] += fftemp[k];
		}

		//draw sigmaf, the field observation error
		rssf=0.0;
#ifdef _OPENMP
#pragma omp parallel private(restemp)
{
		double lrssf=0.0;
#pragma omp for
		for(size_t k=0;k<nf;k++) {restemp=yf[k]-allfitf[k]; lrssf += restemp*restemp;}
#pragma omp atomic
		rssf+=lrssf;
}
#else //_OPENMP
		for(size_t k=0;k<nf;k++) {restemp=yf[k]-allfitf[k]; rssf += restemp*restemp;}
#endif //_OPENMP
		pi.sigmaf = sqrt((nuf*lambdaf + rssf)/gen.chi_square(nuf+nf));

		//draw sigma
		rss=0.0;
#ifdef _OPENMP
#pragma omp parallel private(restemp)
{
		double lrss=0.0;
#pragma omp for
		for(size_t k=0;k<n;k++) {restemp=y[k]-allfit[k]; lrss += restemp*restemp;}
#pragma omp atomic
		rss+=lrss;
}
#else //_OPENMP
		for(size_t k=0;k<n;k++) {restemp=y[k]-allfit[k]; rss += restemp*restemp;}
#endif //_OPENMP
		pi.sigma = sqrt((nu*lambda + rss)/gen.chi_square(nu+n));

		if(i>=burn)
		{
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<n;k++) pmean[k] += allfit[k];
#ifdef _OPENMP
#pragma omp parallel for
#endif
			for(size_t k=0;k<nf;k++) pfmean[k] += allfitf[k];
		}
		if(i>=burn && !(((int)i)%((int)thin)))
		{
			cout << "i=" << i << "...saving draw" << endl;
			bsdf << pi.sigma << endl;
			bsdff << pi.sigmaf << endl;
			for(size_t j=0;j<pi.ncaldims;j++)
				bcpf << pi.calests[j] << " ";
			bcpf << endl;
			for(size_t j=0;j<m;j++) treef << t[j] << endl;  //save all the trees from this draw
		}
		if((i%200)==0)
		{
			cout << "i=" << i << endl;	
			cout << "Tree acceptance rate was: " << (double)(pi.taccept)/(pi.tproposal) << endl;
			cout << "Average tree depth was: " << (double)(pi.tavgd)/(m*200.0) << endl;
			cout << "Maximum tree depth was: " << pi.tmaxd << endl;
			pi.taccept=0;
			pi.tproposal=0;
			pi.tavgd=0;
			pi.tmaxd=0;
		}
	} // end of mcmc loop


	// save in-sample mean prediction of simulator
	for(size_t k=0; k<n; k++) pmean[k] /= (double)nd;
	std::stringstream ofss;
	std::stringstream offss;
	std::string ofs;
	ofss << "bart-fit.txt";
	ofs=ofss.str();
	std::ofstream bfitf(ofs.c_str());
	for(size_t i=0; i<n; i++) bfitf << pmean[i] << endl;

	// save in-sample mean prediction of field
	for(size_t k=0; k<nnf; k++) pfmean[k] /= (double)nd;
	offss << "bart-fit-field.txt";
	ofs=offss.str();
	std::ofstream bfitff(ofs.c_str());
	for(size_t i=0; i<nf; i++) bfitff << pfmean[i] << endl;

	//cleanup
	bfitff.close();
	bfitf.close();
	treef.close();
	bsdf.close();
	bsdff.close();
	delete[] allfit;
	delete[] allfitf;
	delete[] r;
	delete[] rf;
	delete[] ftemp;
	delete[] fftemp;
	delete[] pmean;
	delete[] pfmean;
	delete[] ppredmean;
	delete[] fpredtemp;

	return 0;
}

